/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * COPYING NOTES
 *
 * Socket.cpp -- Implementation of the Socket class.
 * 
 * Copyright (C) 2002, Roberto A. Foglietta <robang@libero.it>
 * Copyright (C) 2002, GEA S.r.l. Generale Elettronica Automotive
 *
 * Copyright (C) 2004, Roberto A. Foglietta <me@roberto.foglietta.name>
 * Copyright (C) 2004, Tecno Elettrica Brianza snc di Riva P. & G.  
 *
 * Copyright (C) 2005, Roberto A. Foglietta <me@roberto.foglietta.name>
 * Copyright (C) 2005, SAD Trasporti Locali Bolzano s.p.a.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * REVISION NOTES:
 * released 21-10-2002 by Roberto A. Foglietta
 * modified 01-08-2005 by Roberto A. Foglietta
 */
 

#include "Socket.h"

#include <errno.h>
#include <fcntl.h>

#include <iostream>


Socket::Socket ():
m_sock (-1)
{
	memset (&m_addr, 0, sizeof (m_addr));
}


Socket::~Socket ()
{
	if (is_valid ())
		::close (m_sock);
}


bool
Socket::create ()
{
	m_sock = socket (AF_INET, SOCK_STREAM, 0);

	if (!is_valid ())
		return false;

	/*
	// TIME_WAIT - argh
	//|SO_KEEPALIVE
	*/
	int on = 1;
	int ret = setsockopt (m_sock, SOL_SOCKET, SO_REUSEADDR, (const char *)&on, sizeof (on));
	if (ret == -1)
		return false;
	
	return true;
}


bool
Socket::bind (const int port)
{
	if (!is_valid ()) {
		return false;
	}

	memset((void *)&m_addr, 0, sizeof(m_addr));
	m_addr.sin_family = AF_INET;
	m_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	m_addr.sin_port = htons (port);

	int bind_return =::bind (m_sock, (struct sockaddr *)&m_addr, sizeof (m_addr));
	if (bind_return < 0) {
		return false;
	}

	return true;
}


bool
Socket::listen () const
{
	if (!is_valid ()) {
		return false;
	}

	int listen_return =::listen (m_sock, MAXCONNECTIONS);
	if (listen_return < 0) {
		return false;
	}

	return true;
}


bool
Socket::accept (Socket & sock) const
{
	if (!is_valid ()) {
		return false;
	}

	int addr_length = sizeof (m_addr);
	sock.m_sock =::accept (m_sock, (struct sockaddr *)&m_addr, (socklen_t *)&addr_length);
	if (sock.m_sock < 0)
		return false;

	return true;
}


bool
Socket::close (Socket & sock) const
//Socket::close () const
{
	if(sock.is_valid()) {
		if(::close(sock.m_sock) != 0) {
	//if(is_valid()) {
	//	if(::close(m_sock) != 0) {
			return false;
		}
	}
	
	return true;
}


bool
Socket::send (const std::string s) const
{
	int status =::send (m_sock, s.c_str (), s.size (), MSG_NOSIGNAL);
	if (status == -1)
		return false;
	
	return true;
}


int
Socket::recv (std::string & s) const
{
	char buf[MAXRECV + 1];
	int length;

	s = "";
	if(!is_ready())
		return -1;

	memset (buf, 0, MAXRECV + 1);	
	length =::recv (m_sock, buf, MAXRECV, 0);
	if (length > 0)
		s = buf;
	
	//std::cerr << "errno = " << errno << " of " << EINTR << " length = " << length << "\n";	
	return length;
}


bool
Socket::connect (const std::string host, const int port)
{
	if (!is_valid ())
		return false;

	m_addr.sin_family = AF_INET;
	m_addr.sin_port = htons (port);

	int status = inet_pton (AF_INET, host.c_str (), &m_addr.sin_addr);

	if (errno == EAFNOSUPPORT)
		return false;

	status =::connect (m_sock, (struct sockaddr *) & m_addr, sizeof (m_addr));

	if (status == 0)
		return true;
	else
		return false;
}


void
Socket::set_non_blocking (const bool b)
{

	//setBlocking(m_sock, !b);

	int opts = fcntl (m_sock, F_GETFL);
	if (opts < 0) {
		std::cerr << "fcntl(m_sock, F_GETFL) error!\n";
		return;
	}

	if (b)
		opts = (opts | O_NONBLOCK);
	else
		opts = (opts & ~O_NONBLOCK);

	if(fcntl (m_sock, F_SETFL, opts) < 0)
		std::cerr << "fcntl(m_sock, F_SETFL, opts) error!\n";
}


int
Socket::timeout (const int msec, int *mask) const
{
	static int timeout = POLL_TIMEOUT_ms;

	if(mask != NULL)
		*mask = 0;

	if(msec > 0) {
		timeout = msec;
		return 0;
	}

	if (!is_valid ())
		return -1;

	struct pollfd ufds = { m_sock, 0xFFFF, 0 };
	int ret = poll(&ufds, 1, timeout);
	if(mask != NULL)
		*mask = ufds.revents;

	if(ret > 0) {
		ret = ufds.revents & (POLLIN|POLLPRI);
	}

	//std::cerr << "timeout: ret = " << ret << " mask = " << ufds.revents << "(" << (ufds.revents&POLLHUP) << ") \n";
	return ret;
}


int  
Socket::is_hangup() const
{ 
	int i;

	if(timeout(-1, &i) > 0) 
		return i&POLLHUP;
	
	return 0;
}
