/*
    This file is part of the Linux driver 
    for the Eltron p310c plastic card printer

    Copyright (C) 2002 Flemming Frandsen <ff-p310c@partyticket.net>
    Based on initial work by Stian Skjelstad <stian@nixia.no>
    Materials (printer+cards+ribbons) provided by http://www.psi.no/
	
    Copyright (C) 2005 Roberto A. Foglietta <roberto.foglietta@sad.it>
    Copyright (C) 2005 Roberto A. Foglietta <me@roberto.foglietta.name>
	
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <termios.h>

#include "cnvcmp.h"
#include "device.h"
#include "rafdebug.h"

int speeds[] = { 9600, 19200, 38400, 57600, -1 };
int spcods[] = {B9600, B19200, B38400, B57600, 0};
int AbortOnAllErrors = 1;
 
static int fser = -1;

#define ACK  	0x06
#define NACK	0x15
#define EOT	0x04

#define ASW_ACK    1
#define ASW_NACK   0
#define ASW_RAW   -1

#define	TIMEOUT_READ 40
#define TAO 50000L

static
char *cmd(const char *c) {
	static char tmp[128];
	int i; 
	tmp[0] = 0;
	for(i = 0; i < 127 && c[i]; i=i) {
		sprintf(tmp, "%s0x%02x ", tmp, (unsigned char)c[i]);
		i++;
	}
	return tmp;
}

int speed_to_idx(char *str) {
	int def = DEF_SPEED_IDX;

	if(str) {
		int i, bauds = atoi(str);
		for (i = MIN_SPEED_IDX; i <= MAX_SPEED_IDX; i++) 
			if (bauds == speeds[i]) 
				return i;
	}
	fprintf(stderr,"Baud rate unknown, setting default (%d)\n", speeds[def]);
	return def;
}

/******************************************************************************/
 
static struct {
    int	cod;
	int fatal;
    char *msg;
  } errors[] = {
    {  -1,	1, "Mechanical Error - Printer"},
    {   1,	1, "Ribbon Broken/Missing"},
    {   2,	1, "Temperature"},
    {   3,	1, "Mechanical Failure"},
    {   4,	1, "Feeder Empty"},
    {   6,	0, "Card not in encoder"},
    {  10,	0, "Invalid command or Parameter"},
    {  11,	0, "Invalid Coordinates (Image placement)"},
    {  12,	0, "Unknown Bar Code Reference"},
    {  13,	0, "Unknown Text/Font Reference"},
    {  14,	0, "Unknown Command"},
    {  20,	0, "Bar Code Data Syntax"},
    {  21,	0, "Text Data Syntax"},
    {  22,	0, "Graphic Data Syntax"},
    {  30,	0, "Graphic Image Initialization Failed"},
    {  31,	0, "Graphic Image Maximum Width Exceeded"},
    {  32,	0, "Graphic Image Maximum Height Exceeded"},
    {  33,	0, "Graphic Image Data Checksum Error"},
    {  34,	1, "Data Transfer Time-out"},
    {  40,	0, "Parameter/Syntax"},
    {  41,	1, "Mag. Encoder Write"},
    {  42,	1, "Mag. Encoder Read/Verify"},
    {  43,	1, "Mag. Encoder Mechanical"},
    {  44,	1, "Mag. Encoder Not Responding"},
    {  45,	1, "Magnetic Stripe Missing / Card Jam"},
    {   0,	0, NULL }
};

# ifndef	TIOS_VMIN
#	define	TIOS_VMIN	0
# endif

# ifndef	TIOS_VTIME
#	define	TIOS_VTIME	10
# endif


static 
struct termios	orig_tios, *p_orig_tios = NULL;


int set_baud(int fout, int ret)	
{
	DEBUG(0);
	struct termios tios;
	int *v;

	if(!fout)
		fout = fser;
	ret = valid_speed_idx(ret);
	PRINTF("set_baud(%d, %d)", fout, ret);
	v = &spcods[ret];
	
	if ( tcgetattr(fout,&orig_tios)<0 )
		perror("set_baud(): ERROR in configuring device parameters");
	
	tios = orig_tios;
	tios.c_iflag = 0;
	tios.c_oflag = 0;
	tios.c_lflag = 0;
	tios.c_cflag = 0;
	tios.c_iflag |= IGNBRK|IGNCR;
	tios.c_cflag = CS8 | CREAD | HUPCL;
	if ( cfsetospeed(&tios,*v)<0 || cfsetispeed(&tios,*v)<0 ) 
		perror("set_baud(): ERROR in configuring device speed");
	
	tios.c_cc[VMIN] = TIOS_VMIN;
	tios.c_cc[VTIME] = TIOS_VTIME;
	if ( tcsetattr(fout,TCSANOW,&tios)<0 )
		perror("set_baud(): ERROR in resetting device");
	
	p_orig_tios = &orig_tios;
	return ret;
}


void set_parity(int fout, char *sb)
{
	char **s;
	struct termios tios;
	static char *pa[]={"none", "odd", "even", NULL};

	if(!fout)
		fout = fser;

	if ( tcgetattr(fout,&tios)<0 )
		perror("set_baud(): ERROR in configuring device parameters");
	
	for (s=pa; *s; ++s) 
		if (!strcmp(*s,sb)) 
			break;
	if (!(*s)) {
		perror("set_baud(): ERROR in unknown parity, setting default");
		*s = pa[0];
	}
	tios.c_cflag &= ~(PARENB|PARODD);
	
	if ( !strcmp(*s,"odd") )
		tios.c_cflag |= PARENB|PARODD;
	else if ( !strcmp(*s,"even") )
		tios.c_cflag |= PARENB;
	
	if ( tcsetattr(fout,TCSANOW,&tios)<0 )
		perror("set_baud(): ERROR in resetting device");
}



/*
 * Torna : 
 *     1 <- letto ACK
 *	   0 <- letto NAK
 *	  -1 <- altro (may be raw data)
 */
static 
int getAnswer(int fout, char *b, int sz_buf, int *nc)
{
  DEBUG(0);

  int i=0, err=0, start=0, ok=ASW_RAW;
  struct timeval timeout;
  fd_set mask, readfs;

  timeout.tv_sec = TIMEOUT_READ;
  timeout.tv_usec = 0;
  FD_ZERO(&mask);
  FD_SET(fout,&mask);
  readfs = mask;
  for (	readfs=mask;
        select(fout+1,&readfs,(fd_set*)NULL,(fd_set*)NULL,&timeout)>0;
        readfs=mask
	){
    int rr;
    char ch;

    if ( (rr=read(fout,&ch,1))==1 ) {
      int	c = ch &0377;

      PRINTF("RECV_CHAR s=%2d, i=%2d | 0%03o | 0x%02x | %c   ",start,i,c,c,c);
      if (c == ACK) {
        if (ok == ASW_RAW) {
			ok = ASW_ACK;
			start = i;
		}
        break;
      }
  
      if (c == NACK) {
		ok = ASW_NACK;
		start = i;
	  }
      else if ( ok!=ASW_RAW && c==EOT )
        break; 
      else if ( i==sz_buf-1 ) {
		/*
		 * Overflow sul buffer; continua ugualmente
		 * a leggere fino al termine ma non scrive sul buffer.
		 */
        fprintf(stderr,"Overflow reading buffer\n");
        err = 1;
      } else if ( !err ) {
        ++i;
        *b++ = ch;
      }
    }
  }
  b[i] = '\0';	
  *nc = start;
  return ok;
}


/******************************************************************************/

#define FALSE 0
#define TRUE  1

static
int tpause(int timeout) {
	DEBUG(0);
	static int count = 0; 		
	//PRINTF("count %d timeout %d\n", count, timeout);

	if(timeout) {
		usleep(TAO);	
		if((++count) > ((timeout<0)?-timeout:timeout))	{
			if(timeout < 0) {
				fprintf(stderr,"FATAL ERROR: Writing Timeout Expired\n"); 
				exit(timeout);
			} else {
				fprintf(stderr,"Writing Timeout Expired\n"); 
				return TRUE;
			}
		} else {
			if(!(count%(1000000/TAO)))
				PRINTF("%d/%d\n", count, timeout);
		}
	} else {
		count = 0;
	}

	return FALSE;
}


static
int show_error(FILE *fout, int err) 
{
	int i;
	char *msg = {"Unknown error"};
	for (i = 0; errors[i].msg; i++)
		if(errors[i].cod == err) {
			msg = errors[i].msg;
			break;
		}
		
	fprintf(fout,"Printer return NACK (%d): '%s'\n", err, msg);
	
	return i;
}


static
int error_manager(int fd, int err) 
{
/*
	int i;
	for (i = 0; errors[i].msg; i++)
		if(errors[i].cod == err)
			break;
*/	
	int i = show_error(stderr, err);

	if(errors[i].msg) {
		if(errors[i].fatal) {
			fprintf(stderr,"FATAL ERROR!\n\n");
			exit(-12);
		} else 
		if(AbortOnAllErrors) {
			fprintf(stderr,"ABORTING ON ERROR!\n\n");
			exit(-10);
		}
		
		//if(errors[i].cod > 40)
		//	return sendcmd(fd, "R");
	}

	return -1;
}


static
int check_answer(int fd, const char *comd) {
	DEBUG(0);
	char buf[1024];
	int s, err;

	PRINTF("check_answer(%d, '%s')", fd, comd);

	if(fser <= 0)
		return 0;
	memset(buf, 0, 1024);

	switch(getAnswer(fser, buf, 1024, &s)) {
		case ASW_ACK:
			PRINTF("Printer return ACK: '%s'\n", cmd(buf));
			if(!buf[0])
				break;
			if(!strncmp(comd,"&E",2) || strstr(comd,"[&E")
			|| !strncmp(comd,"&L",2) || strstr(comd,"[&L")) {
				fprintf(stderr, "Magnetic strip: '%s'\n", buf);
			} else
			if(!strncmp(comd,"!V",2) || strstr(comd,"[!V")) {
				fprintf(stderr, "Printer session: '%s'\n", buf);
			} else 
			if(comd[0] == 'V' || strstr(comd,"[V")) {
				fprintf(stderr, "Printer version: '%s'\n", buf);
			} else
			if(strstr(comd,"&CDEW") || strstr(comd,"&CDER")) {
				fprintf(stderr, "Encoder reset: '%s'\n", buf);
			} else 
			if(comd[0] == 'R') {
				fprintf(stderr, "Printer reset: '%s'\n", buf);
			} else 
				fprintf(stderr, "Unknown data: '%s'\n", buf);
			break;
		
		case ASW_NACK:
			err = atoi(&buf[s]);
			//show_error(stderr, err);
			return error_manager(fd, err);
			break;
				
		default:
			fprintf(stderr,"Unknown print failure '%s':\n%s  \n", cmd(buf), buf);
			return 2;
	}	
	
	return 0;	
}


static
void mwrite(int fd, const char *c, int s) {
    DEBUG(0);
    int w, k = 0;

    tpause(TORESET);
    while (s-k > 0) {
	PRINTF("s = %d, s-k = %d\n", s, s-k);
	w = write(fd, &c[k], s-k);

	PRINTF("wrote %d\n", w);
	if (w < 0) {
		if (errno == EINTR 
		||  errno == EAGAIN) {
			if(tpause(TIMEOUT))
				exit(-111);
			continue;
		}
		perror("mwrite()");
		PRINTF("fd = %d\n", fd);
		exit(-61);
	} else {
		tpause(TORESET);
		k += w;
	}
    }
}

#define getargs(cmd, arg) do{						    		\
	va_list ap;													\
	va_start(ap, arg);											\
	vsnprintf(cmd, sizeof(cmd), arg, ap);						\
	va_end(ap);													\
}while(0)

// Send a simple command to the printer.
int sendcmd(int fd, const char *c, ...) {
	DEBUG(0);
	char cmd[1024]; 
	getargs(cmd, c);

	PRINTF("sendcmd(%d, '%s')\n", fd, cmd);
		
	mwrite(fd, "\033", 1);
    mwrite(fd, cmd, strlen(cmd));
	mwrite(fd, "\015", 1);

	return check_answer(fd, cmd); 
}

//static 
//char *colorname[4] = {"Yellow", "Magenta", "Cyano", "Soft Black"};

extern int Compression;

// Send a binary (imagedata-loading) command to the printer:
int sendbmp(int fd, char *data, int *size, const char *cmd, int channel) {

	char *conv = NULL;
	DEBUG(0);
	PRINTF("sendbinary(%d, 0x%04lx, %d, '%s')\n", fd, (long)data, *size, cmd);

	conv = data;
	if(channel == RESIN || channel == VARNISH)
		conv = dataconvert(data, size);
	
    mwrite(fd, "\033", 1);
	mwrite(fd, cmd, strlen(cmd));
	mwrite(fd, conv, *size);
    mwrite(fd, "\015", 1);
	
	if(conv != data && conv)
		free(conv);

	return check_answer(fd, cmd);
}

int sendzbmp(int fd, char *data, int *size, const char *cmd, int channel) {

	char *comp = NULL;
	int ret;

	if (channel == RESIN || channel == VARNISH)
		comp = rlecompress(data, size, EIGHT_BIT_LEVEL);
	else
		comp = rlecompress(data, size, FIVE_BIT_LEVEL);
	ret = sendbmp(fd, comp, size, cmd, channel);
	if(comp)
		free(comp);
	
	return ret;
}



int sendtesto(int fd, char *txt) {
	int err;
	char *row = NULL;

	if(!txt)
		return -1;

	err = 0;
	row = strtok(txt, "\n");
	while(row) {
		err += sendcmd(fd, row);
		row = strtok(NULL, "\n");
	}
	return err;
}

//extern void sendicmd();
//#define setspeed(a,b) sendicmd((a), "+B %d", (b));

int setserial(int fd, char *ser, int sidx, char *prty) {
	DEBUG(0);

	PRINTF("setserial(%d, 0x%04lx, %d, %s)", fd, (long)ser, sidx, prty);
	sidx = valid_speed_idx(sidx);

	if(fser > 0) {
		close(fser);
		fser = -1;
	}

	if(ser != NULL) {	
		fser = open(ser, O_RDWR | O_NDELAY  | O_NOCTTY);
		if (fser < 0) {	
			fprintf(stderr, "Error while trying to open %s: %s", ser, strerror(errno));
			return -1;
		}
		set_parity(fser, prty);			
		set_baud(fser, sidx);		
		
		if(fd > 0) {
			//
			if(sendcmd(fd, "+B %d", sidx)) {		
				fprintf(stderr, "Setting serial speed fail, continuing without serial\n");
				close(fser);
				fser = -1;
				return 1;			
			}
			/*
			char tmp[32];
			sprintf(tmp, "\033%s%d\015", SET_SPEED, sidx);	
			mwrite(fd, tmp, strlen(tmp));
			PRINTF(cmd(tmp));
		
			if(check_answer(fd, SET_SPEED)) {
				fprintf(stderr, "Setting serial speed fail\n");
				//close(fser);
				//fser = -1;
				return 1;
			}*/	
		} else {
			fprintf(stderr, "No valid file descriptor\n");
			exit(-25);
		}
	} else {
		int i;
		char tmp[32];

		fser = fd;
		set_parity(fser, prty);			
		sprintf(tmp, "\033%s%d\015", SET_SPEED, sidx);	
		for (i = MIN_SPEED_IDX; i <= MAX_SPEED_IDX; i++) {
			set_baud(fser, i);
			mwrite(fser, tmp, strlen(tmp));
			PRINTF(cmd(tmp));
			usleep(TAO);
		}
		set_baud(fser, sidx);
		//sprintf(tmp, "\033E\015");	
		//PRINTF(cmd(tmp));
		//mwrite(fser, tmp, strlen(tmp));
	
		if(check_answer(fd, SET_SPEED)) {
			fprintf(stderr, "Setting serial speed fail and the unique device available is useless!\n");
			exit(-129);
		}	
	}
	
	return 0;

	/*
	if(!ser) { 
		fprintf(stderr,"ERROR: setting serial file descriptor failed\n");
		exit(-114);
	} 
	*/

	/*
	if(fser > 0)
		close(fser);
	
	fser = open(ser, O_RDWR | O_NDELAY  | O_NOCTTY);
	if (fser < 0) {	
		fprintf(stderr, "Error while trying to open %s: %s", ser, strerror(errno));
		exit(-2);
	}

	int i;
	char tmp[32];
	sprintf(tmp, "\033%s%d\015", SET_SPEED, sidx);	
	set_parity(fser, prty);			
	for (i = MIN_SPEED_IDX; i <= MAX_SPEED_IDX; i++) {
		set_baud(fser, i);
		mwrite(fser, tmp, strlen(tmp));
		PRINTF(cmd(tmp));
		usleep(TAO);
	}
	set_baud(fser, sidx);
	//sprintf(tmp, "\033E\015");	
	//PRINTF(cmd(tmp));
	//mwrite(fser, tmp, strlen(tmp));

	if(check_answer(SET_SPEED)) {
		fprintf(stderr, "Setting serial speed fail\n");
		exit(-129);
	}
	return fser;
	*/

/*******************************************************************************
	set_baud(fser, MIN_SPEED_IDX);
	set_parity(fser, "none");
	
	sidx = valid_speed_idx(sidx);
	if(sidx != MIN_SPEED_IDX) {
		char tmp[32];
		sprintf(tmp, "\033%s%d\015", SET_SPEED, sidx);	
		PRINTF(cmd(tmp));
		mwrite(fser, tmp, strlen(tmp));
		set_baud(fser, sidx);
		set_parity(fser, "none");			
	
		//sprintf(tmp, "\033E\015");	
		//PRINTF(cmd(tmp));
		//mwrite(fser, tmp, strlen(tmp));
	
		if(!check_answer(SET_SPEED)) {
			break;
			//fprintf(stderr, "Setting serial speed fail\n");
			//set_baud(fser, MIN_SPEED_IDX);
		}
	}
	return fser;
*******************************************************************************/
}


struct {
	int code;
	char *str;
} configure[] = {
	{ 0, "Black Printing Parameters" },
	{ 1, "X Offset" },
	{ 2, "Y Offset" }, 
	{ 3, "Black Contrast" },
	{ 4, "Varnish Contrast" },
	{ 5, "Hologram Contrast" },
	{ 6, "Yellow Contrast" },
	{ 7, "Magenta Contrast" },
	{ 8, "Cyan Contrast" },
	{ 9, "Kdye Intensity"},
	{10, "Yellow Intensity" },
	{11, "Magenta Intensity" },
	{12, "Cyan Intensity" },
	{13, "Kdye Intensity"},
	{14, "SXY Origin Offset (0=yes,1=no)"},
	{15, "Print Head Resistance" },
	{16, "Black Speed" },
	{17, "Varnish Speed" },
	{18, "+EC Command dots" },
	{19, "Smart Card Offset"},
	{20, "Magnetic Encoder (0=no,1=yes)" },
	{21, "Coercivity Setting (0=lo,1=hi)" }, 
	{22, "Magnetic Encoding Format (0=JIS2,1=ISO)" }, 
	{23, "Encoder Head Placement (0=Below,1=Above) Card Path" }
};


int query_printer(int fd) {
	int i;
	
	for(i = 0; i < 24; i++) {
		fprintf(stderr,"\n%s: ",configure[i].str);
		if(sendcmd(fd, "m 1 .[!V %d", i))
			;
			//return i+1;
	}
	//sendcmd(fd, "!V");
	return 0;
}
