/*
  This file is part of the Linux driver for the Eltron p310c plastic card printer
	Copyright (C) 2002 Flemming Frandsen <ff-p310c@partyticket.net>
	Based on initial work by Stian Skjelstad <stian@nixia.no>
	Materials (printer+cards+ribbons) provided by http://www.psi.no/
	
	Copyright (C) 2005, Roberto A. Foglietta <roberto.foglietta@sad.it> 
	Copyright (C) 2005, Roberto A. Foglietta <me@roberto.foglietta.name>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
*/

#include "images.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>


CMYKrImage *new_CMYKrImage(CMYKrImage *output) {
	if(!output) 
		output = (CMYKrImage *)malloc(sizeof(CMYKrImage));

	output->height =  0;
	output->width  =  0;
	output->color[0] = NULL;
	output->color[1] = NULL;
	output->color[2] = NULL;
	output->color[3] = NULL;

	return output;
}

void copyImg(Image src, Image dst, int x0, int y0) {
    Pixel *source = src.imagedata;
    Pixel *target = dst.imagedata+y0*dst.width+x0;
    int y, ll = sizeof(int)*src.width;

    for (y = 0; y<src.height; ++y) {
		memcpy(target,source,ll);
		target += dst.width;
		source += src.width;
    }
}

Image makeImage(int width, int height) {
	int bytes;
    Image img;
    img.height = height;
    img.width = width;
    bytes = sizeof(int)*img.width*img.height;
    img.imagedata = (unsigned int *)malloc(bytes);    
    memset(img.imagedata,0,bytes);
    return img;
}


CMYKrImage makeCMYKrImage(int width, int height) {
	int color, size;
    CMYKrImage cmy;

    cmy.width = width;
    cmy.height = height;
	size = width*height;
    for (color=0;color < 4; color++) 
		cmy.color[color] = (unsigned char *)malloc(size);    
	
    return cmy;
}

CMYKrImage *newCMYKrImage(int width, int height) {
    CMYKrImage *img = (CMYKrImage *)malloc(sizeof(CMYKrImage));
    *img = makeCMYKrImage(width,height);
    return img;
}


#define TRUE  1
#define FALSE 0
int has_alpha(Image *overlay, int size) {
	int pixel;

	if(!overlay)
		return FALSE;
	
	for (pixel = 0; pixel < size; pixel++) {
		unsigned int opixel = overlay->imagedata[pixel];
		int alpha = opixel >> 24;
		if(alpha)
			return TRUE;
	}
	return FALSE;
}


#define C(a) (255-(((a) >> 16) & 0xff))
#define M(a) (255-(((a) >> 8)  & 0xff))
#define Y(a) (255-((a) & 0xff))
#define K(a) (!((a)&0x00ffffff))

CMYKrImage rgb2cmykr(Image *color, Image *black) {

	Image *p; 
	CMYKrImage cmy;
	int line, size, pixel;

	if(!color && !black) {
		fprintf(stderr,"Error: one layer has to be defined, at least!\n");
		exit(-4);
	}

    if (black && color) {
		if (color->width  != black->width 
		||	color->height != black->height) {
			fprintf(stderr,"Error: The color and black layer has not the same size (%d,%d) != (%d,%d)\n",
					color->width, color->height, black->width, black->height);
			exit(-4);
		}
    }

	p = (color)?color:black; 
	line = p->height*sizeof(unsigned char);
	size = p->width*line;
	cmy = makeCMYKrImage(p->width,p->height);

    for (pixel = 0; pixel < size; pixel++) {
		// Start out by letting the bottom / background set the output color:
		unsigned int cyan    = 0;
		unsigned int magenta = 0;
		unsigned int yellow  = 0;
		unsigned char resin  = 0;
	
		if(color) {
			cyan    = C(color->imagedata[pixel]);
			magenta = M(color->imagedata[pixel]);
			yellow  = Y(color->imagedata[pixel]);
		}

		if (black)
			resin = K(black->imagedata[pixel]);
		
		cmy.color[CYAN][pixel]    = cyan >> 3;
		cmy.color[MAGENTA][pixel] = magenta >> 3;
		cmy.color[YELLOW][pixel]  = yellow >> 3;
		cmy.color[BLACK][pixel]   = resin;
		
	    //fprintf(stderr,"pixel: %dx%d; rgb:%06x; cmyk: %d,%d,%d,%d\n",
		//				pixel%line, pixel/line, rgbpixel, cyan,magenta,yellow,resin);
    }

    return cmy;
}

void destroyImage(Image img) {
    free(img.imagedata); 
	img.imagedata = NULL;
}

void destroyCMYKrImage(CMYKrImage cmy) {
	int color;
    for (color=0; color < 4; color++) {
		free(cmy.color[color]);
		cmy.color[color] = NULL;
    }
}

const int MAGIC = 0x0F155E + 1;

CMYKrImage *loadCMYKrImage(char *fn) {
	int color;
	int magic = 0;
	int width = 0;
	int height = 0;
	CMYKrImage *img = NULL;
	FILE *f	 = NULL;		

	if(fn)
		f = fopen(fn, "r");	
	
	if (!f) 
		return NULL;

	if (fread(&magic,sizeof(magic),1,f) == 1 
	&&	magic == MAGIC 
	&&	fread(&width,sizeof(width),1,f) == 1 
	&&	fread(&height,sizeof(height),1,f) == 1 
	&&	width > 0 && height > 0 && width * height < 1000000) {
		img = newCMYKrImage(width, height);
		for (color=0;color < 4; color++) 
			fread(img->color[color],width*height,1,f);
	}

	fclose(f);
    return img;
}

int saveCMYKrImage(char *fn, CMYKrImage *img) {	
	int magic = MAGIC;
	int color;
    FILE *f = fopen(fn, "w");
    if (!f) 
		return false;
	
	fwrite(&magic,sizeof(magic),1,f);
	fwrite(&img->width,sizeof(img->width),1,f);
	fwrite(&img->height,sizeof(img->height),1,f);
	for (color=0;color < 4; color++) 
			fwrite(img->color[color], img->width*img->height,1,f);
	
	fclose(f);
	return true;
}


char *monochrome(CMYKrImage img, int x, int y, int width, int height) {
	int j, xc, yc;
	char *res = NULL;

    if (width % 8) {
		fprintf(stderr, "width of monochrome area is not divisable by 8: %i\n",width);
		exit(-11);
    }
    if (x % 8) {
		fprintf(stderr, "x location of monochrome area is not divisable by 8: %i\n",x);
		exit(-11);
    }

	j = 0;
	res = (char *)malloc(width*height/8);   
    for (yc = 0; yc<height; yc++) {
		int bit = 0;
		char byte = 0;
		for (xc = 0; xc<width; xc++) {
			if (img.color[BLACK][x+xc+img.width*(y+yc)]) {
				byte |= 1 << (7-bit);
			}	
			if (++bit > 7) {
				bit = 0;
				res[j++] = byte;
				byte = 0;
			}
		}
    }

    return res;
}

char *colorrect(CMYKrImage img, int x, int y, int width, int height, int color) {
    int xc, yc, j;
	char *res = NULL;

    if (color < 0 || color > 3) {
		fprintf(stderr,"Invalid color passed to colorrect: %d\n",color); 
		exit(-1);
    }

	j = 0;
	res = (char *)malloc(width*height);
    for (yc = 0; yc<height; yc++) 
		for (xc = 0; xc<width; xc++) 
			res[j++] = img.color[color][(yc+y)*img.width+(xc+x)];

    return res;
}
