/*
  This file is part of the Linux driver for the Eltron p310c plastic card printer
    Copyright (C) 2002 Flemming Frandsen <ff-p310c@partyticket.net>
    Based on initial work by Stian Skjelstad <stian@nixia.no>
    Materials (printer+cards+ribbons) provided by http://www.psi.no/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include "pngloader.h"
#include "dirty.h"

void mwrite(int fd, char *c, int s) {
    while (s) {
	int wrote=write(fd, c, s);
	if (errno == EINTR) {
	    continue;
	} 

	if (wrote<0) {
	    perror("mwrite()");
	    exit(0);
	} 

	s-=wrote;
    }
}

// Send a simple command to the printer.
void sendcmd(int fd, char *c) {
    char tmp[100];
    
    sprintf(tmp, "\033%s\015", c);
    mwrite(fd, tmp, strlen(tmp));
}

// Send a command with an interpolated integer to the printer
void sendcmd(int fd, char *c, int i) {
    char fmt[100];
    sprintf(fmt, "\033%s\015", c);
    
    char tmp[100];
    sprintf(tmp, fmt, i);
    
    mwrite(fd, tmp, strlen(tmp));
}

// Send a binary (imagedata-loading) command to the printer:
void sendbinary(int fd, char *cmd, char *data, int length) {
    mwrite(fd, "\033", 1);
    mwrite(fd, cmd, strlen(cmd));
    mwrite(fd, data, length);
    mwrite(fd, "\015", 1);
}

void sendbinary(int fd, char *cmd, char *data, int length, int i) {
    char tmp[100];
    sprintf(tmp, cmd, i);

    mwrite(fd, "\033", 1);
    mwrite(fd, tmp, strlen(tmp));
    mwrite(fd, data, length);
    mwrite(fd, "\015", 1);
}

void sendcolor(int fd, CMYKrImage output, DirtyMap *dm, int color) {
    if (dm->dirty[color][0][0] < 0) {
	fprintf(stderr, "Sending full color: %d\n",color);
	
	sendcmd(fd, "$FP %d",color);
	sendbinary(fd, "PS %d 32 ", (char *)output.color[color], output.width*output.height, color);

    } else {
	fprintf(stderr, "Sending delta color: %d\n",color);

	for (int x=0; x<DIRTX; x++) {
	    for (int y=0; y<DIRTY; y++) {
		if (dm->dirty[color][x][y]) {
		    int xstart = x*(WIDTH/DIRTX);
		    int ystart = y*(HEIGHT/DIRTY);
		    char *bytes = colorrect(output, xstart, ystart,
					     WIDTH/DIRTX, HEIGHT/DIRTY, color);

		    char cmd[100];
		    sprintf(cmd,"GS %d 32 %d %d %d %d ",
			    color,
			    HEIGHT-ystart-HEIGHT/DIRTY, xstart,
			    HEIGHT/DIRTY,WIDTH/DIRTX);
		    
		    sendbinary(fd, cmd, bytes, (WIDTH/DIRTX)*(HEIGHT/DIRTY));
		    free(bytes);
		}
	    }
	}
    }

    sendcmd(fd, "IS %d",color); // Print the color
}

void sendblack(int fd, CMYKrImage output, DirtyMap *dm) {
    if (dm->dirty[BLACK][0][0] < 0) {
	char *black = monochrome(output, 0,0,WIDTH,HEIGHT);
/*
	char rlebuffer[2*1008*624];
	int compressed_size = rlecompress(black, output.width*output.height/8, rlebuffer);
	fprintf(stderr,"compressed %d bytes to %d bytes\n", output.width*output.height/8, compressed_size);

	fprintf(stderr, "Sending compressed Black\n");
	sendcmd(fd, "F"); 
	sendcmd(fd, "G 0 0 2 78 1008 1"); 
	write(fd, "\033Z", 2);
	mwrite(fd, rlebuffer, compressed_size);
	write(fd, "\015", 1);
	free(black);
*/
	
	fprintf(stderr, "Sending full Black\n");
	sendcmd(fd, "F"); 
	sendcmd(fd, "G 0 0 0 78 1008 1"); 
	sendbinary(fd, "Z", black, output.width*output.height/8);
	free(black);
	
    } else {
	fprintf(stderr, "Sending delta Black\n");

	for (int x=0; x<DIRTX; x++) {
	    for (int y=0; y<DIRTY; y++) {
		if (dm->dirty[BLACK][x][y]) {
		    int xstart = x*(WIDTH/DIRTX);
		    int ystart = y*(HEIGHT/DIRTY);
		    char *black = monochrome(output, xstart, ystart,
					     WIDTH/DIRTX, HEIGHT/DIRTY);

		    char cmd[100];
		    
		    sprintf(cmd,"G %d %d 0 %d %d 1",
			    HEIGHT-ystart-HEIGHT/DIRTY, xstart,
			    WIDTH/DIRTX/8,
			    HEIGHT/DIRTY);
		    
		    sendcmd(fd, cmd); 
		    sendbinary(fd, "Z", black, ((WIDTH/DIRTX)*(HEIGHT/DIRTY)) / 8);
		    free(black);
		}
	    }
	}
    }

    fprintf(stderr, "Printing black\n");
    sendcmd(fd, "I 10");
}

void sendwhite(int fd, CMYKrImage output, DirtyMap *dm) {
    if (dm->dirty[BLACK][0][0] < 0) {
	char *black = monochrome(output, 0,0,WIDTH,HEIGHT);
/*
	char rlebuffer[2*1008*624];
	int compressed_size = rlecompress(black, output.width*output.height/8, rlebuffer);
	fprintf(stderr,"compressed %d bytes to %d bytes\n", output.width*output.height/8, compressed_size);

	fprintf(stderr, "Sending compressed Black\n");
	sendcmd(fd, "F"); 
	sendcmd(fd, "G 0 0 2 78 1008 1"); 
	write(fd, "\033Z", 2);
	mwrite(fd, rlebuffer, compressed_size);
	write(fd, "\015", 1);
	free(black);
*/
	
	fprintf(stderr, "Sending full White\n");
	sendcmd(fd, "vF"); 
	sendcmd(fd, "vG 0 0 0 78 1008 1"); 
	sendbinary(fd, "vZ", black, output.width*output.height/8);
	free(black);
	
    } else {
	fprintf(stderr, "Sending delta White\n");

	for (int x=0; x<DIRTX; x++) {
	    for (int y=0; y<DIRTY; y++) {
		if (dm->dirty[BLACK][x][y]) {
		    int xstart = x*(WIDTH/DIRTX);
		    int ystart = y*(HEIGHT/DIRTY);
		    char *black = monochrome(output, xstart, ystart,
					     WIDTH/DIRTX, HEIGHT/DIRTY);

		    char cmd[100];
		    
		    sprintf(cmd,"vG %d %d 0 %d %d 1",
			    HEIGHT-ystart-HEIGHT/DIRTY, xstart,
			    WIDTH/DIRTX/8,
			    HEIGHT/DIRTY);
		    
		    sendcmd(fd, cmd); 
		    sendbinary(fd, "vZ", black, ((WIDTH/DIRTX)*(HEIGHT/DIRTY)) / 8);
		    free(black);
		}
	    }
	}
    }

    fprintf(stderr, "Printing white\n");
    sendcmd(fd, "I 10");
}

int repeats(char *data, int pos, int length, int limit) {
    int count = 0;
    char value = data[pos];
    while (pos+count < length && count < limit) {
	if (data[pos+count] != value) {
	    return count;
	}
	count++;
    }
}

int rlecompress(char *indata, int inlength, char *outdata) {
    int i=0;
    int j=0;

    while (i < inlength) {
	// If the data repeated more than twice then emmit a 'repeat' code:
	int count = repeats(indata,i,inlength,126);
	if (count > 2 || i==0) {
	    outdata[j++] = 127 || count;
	    outdata[j++] = indata[i];
	    i += count;
	} 

	// See how long the data doesn't repeat.
	count = 0;
	while (i+count < inlength && repeats(indata,i+count,inlength,30) <= 2) {
	    count++;
	}
	if (count) {
	    outdata[j++] = count-1;
	    for (int k=0;k<count-1;k++) {
		outdata[j++] = indata[i+k];
	    }
	    i += count;
	}
    }

    return j;
}

int main(int argc, char *argv[]) {
    char * param_encode_data=NULL;
    char * param_encode_track=NULL;
    char * param_background=NULL;
    char * param_foreground=NULL;
    char * param_device=NULL;
    char * param_state=NULL;
    
    // Parse command line params
    for (int i=0; i < argc-1; i++) {
        if (!strcmp(argv[i],"-b")) {
            param_background=argv[++i];
        } 
        else if (!strcmp(argv[i],"-f")) {
            param_foreground=argv[++i];
        }
        else if (!strcmp(argv[i],"-d")) {
            param_device=argv[++i];
        }
        else if (!strcmp(argv[i],"-s")) {
            param_state=argv[++i];
        }
        else if (!strcmp(argv[i],"-e")) {
            param_encode_data=argv[++i];
        }
        else if (!strcmp(argv[i],"-t")) {
            param_encode_track=argv[++i];
        }
    }

    if (!param_device || !param_background || !param_state) {
	fprintf(stderr,"Invalid syntax\n");
	fprintf(stderr,"usage: max3300_print -d <lp device> -s <state file> -b <background.png> [ -f <foreground.png> -t <trackno> -e <encoding data>]\n");
	exit(-1);	
    }

    if ((param_encode_data && !param_encode_track) || (param_encode_track && !param_encode_data)) {
        fprintf(stderr,"Must specify both a track to write to and the data to write");
        exit(-1);
    }

    // Open the printer and blow up if it doesn't work.
    int fd = open(param_device, O_RDWR);
    if (fd < 0) {	
	fprintf(stderr, "Error while trying to open %s: %s",param_device,strerror(errno));
	exit(-2);
    }
    
    // Load the state file this is how we avoid sending all the data to the printer every time:
    CMYKrImage *state = loadCMYKrImage(param_state);
    if (!state) {
	fprintf(stderr, "Warning: unable to read state file: %s: %s (printing will be slow)\n",
		param_state,strerror(errno));
	
    }

    // Load the background image:
    Image background = png_load(param_background);
    if (background.width != WIDTH ||
	background.height != HEIGHT) {
	
	fprintf(stderr, "Error: the background image was not 624 x 1008 pixels (it was: %d x %d)\n",
		background.width, background.height);
	exit(-5);
    }

    // Calculate the output image in CMYKr
    CMYKrImage output;    
    if (argc == 5) {
	Image foreground = png_load(param_foreground);
	output = rgb2cmykr(background, &foreground);
	destroyImage(background);
	destroyImage(foreground);

    } else {
	output = rgb2cmykr(background, NULL);
	destroyImage(background);
    }
    
    // Now calculate the changed parts of the image:
    DirtyMap *dm = calculateDirtyMap(state, &output); 

    sendcmd(fd, "+RIB21"); // YMCKr_Kr ribbon
    
    sendcmd(fd, "+EC 0"); //Sync ribbons to yellow/start

    // Set contrast:
    sendcmd(fd, "+$C 0 5");
    sendcmd(fd, "+$C 1 5");
    sendcmd(fd, "+$C 2 5");

    sendcolor(fd, output, dm, YELLOW);
    sendcolor(fd, output, dm, MAGENTA);
    sendcolor(fd, output, dm, CYAN);
    sendblack(fd, output, dm);

/*  No varnish on Max3300  
    fprintf(stderr, "Applying varnish\n");
    sendcmd(fd, "MIB"); // Double-force-check that card is in write-ready position
    sendcmd(fd, "IV 1"); 
*/

    // Not sure about this for Max3300:
    if (param_encode_track && param_encode_data) {
        fprintf(stderr, "Encoding magnetic strip\n");
        char encode_cmd[1024];
        snprintf(encode_cmd,1024,"&E%s %s", param_encode_track, param_encode_data);
        sendcmd(fd, encode_cmd);

    }
    sendcmd(fd, "MO"); // Exit clear card 
    sendcmd(fd, "MF"); // Begin white sequence.
    sendwhite(fd, output, dm);
    sendcmd(fd, "MO"); // Exit white card 
 
    
    close(fd);

    if (!saveCMYKrImage(param_state,&output)) {
	fprintf(stderr, "Unable to write printer state to %s",param_state);
    }

    free(dm);
    destroyCMYKrImage(output);

    exit(0);
}
