/*
  This file is part of the Linux driver for the Eltron p310c plastic card printer
    Copyright (C) 2002 Flemming Frandsen <ff-p310c@partyticket.net>
    Based on initial work by Stian Skjelstad <stian@nixia.no>
    Materials (printer+cards+ribbons) provided by http://www.psi.no/
	
	Copyright (C) 2005 Roberto A. Foglietta <roberto.foglietta@sad.it>
    Copyright (C) 2005 Roberto A. Foglietta <me@roberto.foglietta.name>
	
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>

#include "../include/iofile.h"
#include "pngloader.h"
#include "device.h"
#include "dirty.h"
#include "rafdebug.h"

#define FORWARD 0
#define REVERSE 1

#define FALSE 0
#define TRUE  1

#define LOW 0
#define HIGH 1

//#define INIT_PRINTER "m 1 .[+CDOTS 0 1[%%N[V [!V[+BS 1[+EC 8[+RIB 0[+O 10[+OY 30[SXY 0"
//#define INIT_PRINTER "m 1 .[V[+BS 1[+EC 8[+RIB 0[SXY 1"
const char *INITPRINTER={"m 1 .[V[+BS 1[+EC 8[+RIB 0"};
const char *INITPRINTBW={"m 1 .[V[+BS 1[+EC 0[+RIB 4"};
const char *INITENCODER={"m 1 &R[&CDEW 0 0[&CDER 0 0"};
#define INIT_PRINTER ((char *)INITPRINTER)
#define INIT_PRINTBW ((char *)INITPRINTBW)
#define INIT_ENCODER ((char *)INITENCODER)

int NumberCopies =  1;
int Intensity    =  6;		//4 for B/N
int PrintSpeed   =  4;
int ColorGamma   =  2;
int InitPrinter  =  TRUE;
int CorrectDots  = -1;
int XoffsetDots  = -1;
int YoffsetDots  = -1;
int Compression  =  1;
//int Monochrome  =  0;
const char *VERSION = "2006-02-16 10:28";
const char *HTTPSITE = "http://www.roberto.foglietta.name/pub/eltron_p310c \n";
const char *USAGE = "								\n"
"Usage: 									\n"
" p310_print -d <lp device> -k <serial device> 					\n"
"										\n"
"Options which exec & exit:							\n"
" --version									\n"
" --help									\n"
" --reset									\n"
" --eject									\n"
" --query									\n"
" --testcard									\n"
" --syncribbon 									\n"
" --loadribbon <M|C>								\n"
"										\n"
"Options for printing card:							\n"
" -s <previous status filename> (deprecated)					\n"
" -r <color layer .png filename>						\n"
" -b <black layer .png filename>						\n"
" -m <monochrome .png filename> 						\n"
" -c <other commands filename>							\n"
" -z <compression: (on)|off> 							\n"
" -n <number of copies: (1)...50> 						\n"
"										\n"
"Options for setting images:							\n"
" -G <color gamma level: 0(light)-10(dark)>					\n"
"										\n"
"Options for setting printer:							\n"
" -N <initialize printer: (on)|off>						\n"
" -I <intensity value: 0(low)-10(high)>						\n"
" -S <speed index: 0(fast)-10(slow)>						\n"
" -D <correct dots value: 1-50>							\n"
" -X <offset Xdots value: 0-20>							\n"
" -Y <offset Ydots value: 0-60>							\n"
" -E <stop on all errors: (on)|off>						\n"
"										\n"
"Options for encoding data:							\n"
" -t<N. of track to read> 							\n"
" -e<N. of track to write> \"<data to encode>\"					\n"
" -ec <encoding coercivity: high|low>						\n"
" -ed <encoding direction: forward|reverse> 					\n"
"										\n"
"Note: 										\n"
" In both foreground and background images 100% black (RGB: 0,0,0) becomes 	\n"
"resin black, so if you are printing a photo make sure that the blackest 	\n"
"areas are not 0,0,0 but rather 1,0,0 (or some other non-100% black color) 	\n"
"as resin black is much blacker than CMY-black, to have the top layer become	\n" 
"resin black the alpha of the pixel must be 100% opaque.			\n"
"										\n"
"Note: 										\n"
" In printing with color if you see resin black do not align with CMY-black	\n"
"so try to set correct dots value using data/cross.png image            	\n"
"										\n";


int is_changed(DirtyMap *dm) {
	int x, y;
	if(!dm)
		return true;
	for (x=0; x<DIRTX; x++) 
		for (y=0; y<DIRTY; y++) 
			if (dm->dirty[BLACK][x][y]) 
				return true;
	return false;
}

static
void sendresin(int fd, CMYKrImage output, DirtyMap *dm) {
	int total = 0;
	if(is_changed(dm)) {
		char *black = monochrome(output, 0, 0, WIDTH, HEIGHT);
		int size = WIDTH*HEIGHT/8;
	
		if(Compression) {
			fprintf(stdout, "Sending RLE compressed full Black...\n");
			sendcmd(fd,"m 1 vF[F[G 0 0 2 78 1008 1");
			sendzbmp(fd, black, &size, "Z", RESIN);
		} else {
			fprintf(stdout, "Sending full Black...\n");
			sendcmd(fd,"m 1 vF[F[G 0 0 0 78 1008 1");
			sendbmp(fd, black, &size, "Z", RESIN);
		}
		free(black);
		total += size + strlen("m 1 F[G 0 0 x 78 1008 1[Z");
	}
	fprintf(stdout, "Total bytes sent %d, original was %d bytes\n", total, WIDTH*HEIGHT/8+16);
}


//static 
//char *colorname[4] = {"Yellow", "Magenta", "Cyano", "Soft Black"};

void sendcolor(int fd, CMYKrImage output, DirtyMap *dm, int color) {
	int total = 0;
	static char cmd[32];

    if (is_changed(dm)) {
		int size = WIDTH*HEIGHT;
		sendcmd(fd, "$FP %d", color);

		if(Compression) {
			sprintf(cmd,"GS %d 30 0 0 1008 624 ", color);
			//sprintf(cmd,"PS %d 30 ", color);
			sendzbmp(fd, (char *)output.color[color], &size, cmd, color);
		} else {
			sprintf(cmd,"GS %d 32 0 0 1008 624 ", color);
			//sprintf(cmd,"PS %d 32 ", color);
			sendbmp(fd, (char *)output.color[color], &size, cmd, color);
		}
		total += size + strlen("GS x 3z 0 0 1008 624 ");
	}
	fprintf(stdout, "Total bytes sent %d, original was %d bytes\n", total, WIDTH*HEIGHT);
}

#define printerror(x...) do { fprintf(stderr, x); error=1; } while(0)

#define set_exec_n_quit(x) do { if(exec_n_quit) {					\
	printerror("Two or more long options could not be used at the same time"); 	\
	return 8; 									\
	} else exec_n_quit = x; }  while(0)

int main(int argc, char *argv[]) {
	char *param_color	= NULL;
	char *param_black	= NULL;
	char *param_monokr	= NULL;
	char *param_device	= NULL;
	char *param_state	= NULL;
	char *param_serial	= NULL;
	char *param_testo	= NULL;
	char *exec_n_quit	= NULL;
	char *testo		= NULL;
	char *init_printer      = INIT_PRINTBW;
	char *param_encode_data[3]    = {NULL,NULL,NULL};
	int   param_encode_track[3]   = {0, 0, 0};
	int   param_encode_direction  = FORWARD;
	int   param_encode_coercivity = LOW;
	int   param_encode = 0;
	int   param_sspeed = DEF_SPEED_IDX;
	int   i, fd = -1, error = 0;
	
	CMYKrImage output;
	CMYKrImage *state = NULL;
	DirtyMap *dm = NULL; 
	Image color, black;
	
	fprintf(stdout, "Eltron p310c linux driver, v. %s\n%s\n", VERSION, HTTPSITE);
	
	// Parse command line params
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i],"-h")
		||  !strcmp(argv[i],"--help")) {
			fprintf(stdout,"%s\n", USAGE);
			return 0;
		}
		if (!strcmp(argv[i],"-m"))
			Intensity = 4;
	}

	for (i = 1; i < argc; i++) {
		//fprintf(stderr,"Evaluating cli opt: '%s'\n", argv[i]); 
		if (!strcmp(argv[i],"-e1")) {
			param_encode_data[0] = argv[++i];
			param_encode++;
		} else 
		if (!strcmp(argv[i],"-e2")) {
			param_encode_data[1] = argv[++i];
			param_encode++;
		} else 
		if (!strcmp(argv[i],"-e3")) {
			param_encode_data[2] = argv[++i];
			param_encode++;
		} else
		if (!strcmp(argv[i],"-t1")) {
			param_encode_track[0] = TRUE;
			param_encode++;
		} else 
		if (!strcmp(argv[i],"-t2")) {
			param_encode_track[1] = TRUE;
			param_encode++;
		} else 
		if (!strcmp(argv[i],"-t3")) {
			param_encode_track[2] = TRUE;
			param_encode++;
		} else 
		if (!strcmp(argv[i],"-ec")) {
			if(!strcasecmp("high", argv[++i]))
				param_encode_coercivity = HIGH;
		} else 
		if (!strcmp(argv[i],"-ed")) {
			if(!strcasecmp("reverse", argv[++i]))
				param_encode_direction = REVERSE;
		} else
		if (!strcmp(argv[i],"-r")) {
			init_printer = INIT_PRINTER;
			param_color = argv[++i];
        	} else
		if (!strcmp(argv[i],"-b")) {
			param_black = argv[++i];
		} else
		if (!strcmp(argv[i],"-d")) {
			param_device = argv[++i];
		} else
		if (!strcmp(argv[i],"-s")) {
			param_state = argv[++i];
		} else
		if (!strcmp(argv[i],"-k")) {
			char *str = strdup(argv[++i]);
			param_serial = strtok(str,",");
			param_sspeed = speed_to_idx(strtok(NULL,","));
		} else
		if (!strcmp(argv[i],"-z")) {
			char *tmp = argv[++i];
			Compression = strcasecmp(tmp,"off");
		} else
		if (!strcmp(argv[i],"-m")) {
			init_printer = INIT_PRINTBW;
			param_monokr = argv[++i];
		} else
		if (!strcmp(argv[i],"-n")) {
			int tmp = atoi(argv[++i]);
			if(tmp < 1 || tmp > 50)
				printerror("Number of copies %d is out of range (1-50, default 1)\n\n", tmp);
			else
				NumberCopies = tmp;
		} else
		if (!strcmp(argv[i],"-c")) {
			param_testo = argv[++i];
		} else
		if (!strcmp(argv[i],"-I")) {
			int tmp = atoi(argv[++i]);
			if(tmp < 0 || tmp > 10)
				printerror("Intesity value %d is out of range (0-10, default 4)\n\n", tmp);
			else
				Intensity = tmp;
	        } else
		if (!strcmp(argv[i],"-G")) {
			int tmp = atoi(argv[++i]);
			if(tmp < 0 || tmp > 10)
				printerror("Gamma value %d is out of range (0-10, default 4)\n\n", tmp);
			else
				ColorGamma = tmp;
	        } else
		if (!strcmp(argv[i],"-D")) {
			int tmp = atoi(argv[++i]);
			if(tmp < 1 || tmp > 50)
				printerror("Correction dots value %d is out of range (1-50)\n\n", tmp);
			else
				CorrectDots = tmp;
        	} else
		if (!strcmp(argv[i],"-X")) {
			int tmp = atoi(argv[++i]);
			if(tmp < 0 || tmp > 20)
				printerror("Xoffset dots value %d is out of range (0-20)\n\n", tmp);
			else
				XoffsetDots = tmp;
        	} else
		if (!strcmp(argv[i],"-Y")) {
			int tmp = atoi(argv[++i]);
			if(tmp < 0 || tmp > 60)
				printerror("Yoffset dots value %d is out of range (0-60)\n\n", tmp);
			else
				YoffsetDots = tmp;
        	} else
		if (!strcmp(argv[i],"-S")) {
			int tmp = atoi(argv[++i]);
			if(tmp < 0 || tmp > 10)
				printerror("Slow speed value %d is out of range (0-10)\n\n", tmp);
			else
				PrintSpeed = tmp;
        	} else
		if (!strcmp(argv[i],"-N")) {
			InitPrinter = strcasecmp(argv[++i],"off");
        	} else 	
		if (!strcmp(argv[i],"-E")) {
			AbortOnAllErrors = !strcasecmp(argv[++i],"on");
        	} else
		if (!strcmp(argv[i],"--eject")) {
			set_exec_n_quit(strdup("MO"));
		} else
		if (!strcmp(argv[i],"--reset")) {
			set_exec_n_quit(strdup("R"));
		} else
		if (!strcmp(argv[i],"--query")) {
			InitPrinter = 0;
			set_exec_n_quit(strdup("+CDOTS 0 1\n%%N\nV\n!V"));
		} else
		if (!strcmp(argv[i],"--syncribbon")) {
			set_exec_n_quit(strdup("!FF 0"));
		} else
		if (!strcmp(argv[i],"--loadribbon")) {
			char X = ((++i) < argc)?argv[i][0]:'\0';
			switch(X) {
			case 'M':
				fprintf(stdout,"\nPress a key and after this command get executed change the ribbon...\n");
				init_printer = INIT_PRINTBW;
				break;
			case 'C':
				fprintf(stdout,"\nMount the color ribbon, close the head and press a key to continue...\n");
				init_printer = INIT_PRINTER;
				break;
			default:
				printerror("ERROR: --loadribbon requires a parameter <Monochrome|Color>\n\n");
				return 1;
			}
			getchar();
			exec_n_quit = strdup("MO");
			InitPrinter = TRUE;
		} else
		if (!strcmp(argv[i],"--version")) {
			InitPrinter = 0;
			set_exec_n_quit(strdup("V"));
		} else
		if (!strcmp(argv[i],"--testcard")) {
			set_exec_n_quit(strdup("A 1"));
		} else {
			printerror("\nInvalid option: '%s'\nUse --help to obtain help\n\n", argv[i]);
			return 2;
		}
	}

	if(!param_serial) {
		for (i = 0; i < argc-1; i++) {
			if (!strcmp(argv[i],"--query")) {
				printerror("ERROR: it is possible to query printer only by serial device!\n\n");
				return -1;
			}
		}
	}

	if (!param_device && param_serial) {
		param_device = strdup(param_serial);
		fprintf(stdout,"Using serial port %s at %d bauds as device\n\n", param_device, speeds[param_sspeed]);
	}

	if (!param_device) {
		printerror("\nInvalid syntax, use --help to obtain help\n\n");
		return 3;
	}

	fd = open(param_device, O_RDWR | O_NDELAY  | O_NOCTTY);
	if (fd < 0) {
		printerror("Error while trying to open %s: %s\n\n", param_device, strerror(errno));
		return -2;
	}

	if(param_serial) {
		char *p = NULL;
		if(strcmp(param_device,param_serial))
			p = param_serial; //serial != dev
		setserial(fd, p, param_sspeed, "none");
	}
	
	if(InitPrinter && init_printer) {
		fprintf(stderr, "Initializing system...\n");
		sendcmd(fd, init_printer);	
	}

	if(exec_n_quit) {
		fprintf(stderr, "Executing command...\n");
		char *t = strtok(exec_n_quit, "\n");
		while(t) {
			sendcmd(fd, t);
			t = strtok(NULL, "\n");
		}
		fprintf(stderr,"Command end sucessfully\n");
		close(fd);
		return 0;
	}

	if(param_testo) {
		long lun;
		testo = (char *)read_file_to_buffer (param_testo, "rt", &lun);
		if(!testo) {
			printerror("Error while trying to open %s: %s\n", param_testo, strerror(errno));
			return -3;
		}
	}
	
	/**************************************************************************/

	if(param_state)
		if(!(state = loadCMYKrImage(param_state)))
			printerror("Warning: unable to read state file: %s: %s (printing will be slow)\n", param_state, strerror(errno));

	if(error && AbortOnAllErrors) {
		fprintf(stdout, "Abort on all errors requested and executed\n");
		return 4;
	}

	if((param_color || param_black) && param_monokr) {
		printerror("Error: printing monochrome layer implies you have Kr ribbon and you cannot print black/color layers!\n");
		return -4;
	}
		
	if(param_color || param_black || param_monokr) {
		// Load the background image:
		if (param_monokr) {
			black = png_load(param_monokr, WIDTH, HEIGHT, 1, ColorGamma);
/*			if(black.width  != WIDTH 
			|| black.height != HEIGHT) {
				printerror("Error: the monochrome image is not 624 x 1008 pixels (it is: %d x %d)\n",
												black.width, black.height);
				return -5;
			}
*/		} else {
			// Calculate the output image in CMYKr
			if(param_color) {
				color = png_load(param_color, WIDTH, HEIGHT, 1, ColorGamma);
/*				if(color.width  != WIDTH 
				|| color.height != HEIGHT) {
					printerror("Error: the color image is not 624 x 1008 pixels (it is: %d x %d)\n",
												color.width, color.height);
					return -6;
				}
*/			}
		
			if (param_black) {
				black = png_load(param_black, WIDTH, HEIGHT, 1, ColorGamma);
/*				if(black.width  != WIDTH 
				|| black.height != HEIGHT) {
					printerror("Error: the black image is not 624 x 1008 pixels (it is: %d x %d)\n", 
												black.width, black.height);
					return -7;
				}
*/			}
		}
	
		output = rgb2cmykr((param_color)?&color:NULL, (param_black||param_monokr)?&black:NULL);
	
		if(param_color)
			destroyImage(color);
		if(param_black || param_monokr)
			destroyImage(black);
		if(state)
			dm = calculateDirtyMap(state, &output);
	}

	if(!state && InitPrinter) {
		int n = Intensity, m = PrintSpeed;
		fprintf(stderr, "Initializing printer...\n");

		if(param_color || param_black) {
		//	sendcmd(fd, INIT_PRINTER);	
			sendcmd(fd, "m 1 +$C 0 %d[+$C 1 %d[+$C 2 %d",   n, n, n);
			sendcmd(fd, "m 1 +$L 0 %d[+$L 1 %d[+$L 2 %d",   n, n, n);
		} //else {
		//	sendcmd(fd, INIT_PRINTBW);	
		//}
		sendcmd(fd, "m 1 +V %d[+VC %d[+C %d[+CV %d", m, m, n, n);
		
		if(CorrectDots > 0)
			sendcmd(fd, "+CDOTS %d", CorrectDots);
		if(XoffsetDots > 0)
			sendcmd(fd, "+O %d", XoffsetDots);
		if(YoffsetDots > 0)
			sendcmd(fd, "+OY %d", YoffsetDots);
	}

	if (param_encode) {
		int i, ok;
		char cmd[1024];

		if(!state && InitPrinter) {
			fprintf(stderr, "Initializing encoder...\n");
			sendcmd(fd, INIT_ENCODER);
		}
	
		ok = 0;
		if(InitPrinter)
			sprintf(cmd, "m 1 &C %d[&W %d", param_encode_coercivity, param_encode_direction);
		else
			sprintf(cmd, "m 1 ");
		for(i = 0; i < 3; i++) 
			if(param_encode_data[i])
				ok = sprintf(cmd, "%s[&B %d %s", cmd, i+1, param_encode_data[i]);
		if(ok) {
			fprintf(stderr, "Encoding magnetic strip...\n");
			sendcmd(fd, "%s[&E*", cmd);
		}

		ok = 0;
		if(InitPrinter)
			sprintf(cmd, "m 1 &C %d[&W %d", param_encode_coercivity, param_encode_direction);
		else
			sprintf(cmd, "m 1 ");
		for(i = 0; i < 3; i++)
			if(param_encode_track[i])
				ok = sprintf(cmd, "%s[&L%d", cmd, i+1);
		if(ok)	{
			fprintf(stderr, "Reading magnetic strip...\n");
			sendcmd(fd, cmd);
		}
	}

	if(param_color || param_black) {
		fprintf(stderr, "Flushing color/resin layer...\n");
		for(i = YELLOW; i <= CYAN; i++)
			sendcolor(fd, output, dm, i);
		sendresin(fd, output, dm);
	} else
	if(param_monokr) {
		fprintf(stderr, "Flushing mono/resin layer...\n");
		sendresin(fd, output, dm);
	} else
	if(testo) {
		fprintf(stderr, "Flushing text/resin layer...\n");
		sendcmd(fd, "m 1 F");
	}
		
	if(testo) {
		sendtesto(fd, testo);
	}

	if(param_monokr || param_color || param_black || testo) {
		for(i = 0; i < NumberCopies; i++) {
			if(param_color || param_black) {
				fprintf(stderr, "Printing colors...\n");
				sendcmd(fd, "m 1 IS 0[IS 1[IS 2");
			}
	
			fprintf(stderr, "Printing black...\n");
			sendcmd(fd, "I%s", (param_color || param_black)?" 10":"");
		
			if(param_color || param_black) {
				fprintf(stderr, "Applying varnish...\n");
				sendcmd(fd, "IV 1");
			}
		}
	}
	close(fd);

	if(param_color || param_black || param_monokr) {
		if(param_state)
			if (!saveCMYKrImage(param_state, &output))
				fprintf(stderr, "Unable to write printer state to %s\n", param_state);
		destroyCMYKrImage(output);
	}

	if(dm) 
		free(dm);
	if(testo)
		free(testo);
	
	fprintf(stderr,"Command end sucessfully\n");
	printf("\n");
	
	return 0;
}
